==> newsstar-1.2.2.tar.gz: newsstar-1.2.2/src/group.c <==
/*
    newsstar - advanced news fetcher
    Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* $Id: group.c,v 1.6 2004/08/13 21:48:20 realh Exp $ */

#include "common.h"

#include "group.h"
#include "history.h"
#include "main.h"
#include "server.h"
#include "stats.h"

GroupData *group_new(const char *name, long int last_downloaded,
	ServerData *serv)
{
    GroupData *group = XMALLOC(sizeof(GroupData));

    deblog_info(_("Creating group %s on %s, last downloaded %ld\n"),
	    name, serv->identifier, last_downloaded);
    group->name = XSTRDUP(name);
    group->last_downloaded = last_downloaded;
    group->first_available = group->last_available = 0;
    group->last_requested = -1;
    group->status = group_Virginal;
    group->artnums = NULL;
    group->nthreads = 0;
    group->serv = serv;
    group->bfgroup = NULL;
    group->gc = groupcount_add_group(name);
    return group;
}

void group_delete(GroupData *group)
{
    deblog_debug(_("Deleting group %s\n"), group->name);
    free(group->artnums);
    free(group->name);
    free(group);
}

BOOL group_got_stats(GroupData *group, long int first, long int last,
	int thread_idx)
{
    int i;
    long int new_arts;

    deblog_debug(_("Got stats for %s\n"), group->name);
    group->first_available = first;
    group->last_available = last;
    if (group->last_downloaded >= last)
    {
	group->status = group_AllDone;
	if (group->last_downloaded > last)
	{
	    deblog_warning(_("Last downloaded > last available in %s on %s\n"),
		    group->name, group->serv->identifier);
	    group->last_downloaded = last;
	}
	else
	{
	    deblog_message_thread(thread_idx, _("No new articles in %s\n"),
		    group->name);
	}
	return FALSE;
    }
    group->status = group_Downloading;
    if (group->last_downloaded < 0)
	group->last_downloaded = last + group->last_downloaded;
    if (group->last_downloaded < first)
	group->last_downloaded = first - 1;
    group->first_available = group->last_downloaded + 1;
    group->last_requested = group->first_available - 1;
    new_arts = last - group->first_available + 1;
    /* If this is a new group we may have slipped through above test for
     * 0 new articles so check again */
    if (!new_arts)
    {
	deblog_message_thread(thread_idx, _("No new articles in %s\n"),
		group->name);
	return FALSE;
    }
    deblog_message_thread(thread_idx, _("%ld new %s in %s\n"), new_arts,
	    new_arts == 1 ?  _("article") : _("articles"),
	    group->name);
    stats_inc_total_seen(new_arts);
    group->artnums = XMALLOC(new_arts * sizeof(group_artnum));
    for (i = 0; i < new_arts; ++i)
    {
	group->artnums[i].status = artnum_Unknown;
	group->artnums[i].msgid = NULL;
    }
    return TRUE;
}

static BOOL group_add_or_kill_message_id(GroupData *group, long int artnum,
	const char *msgid, BOOL kill)
{
    static long int hist_dups = 0;
    group_artnum *artinfo = &group->artnums[artnum - group->first_available];
    artinfo->msgid = hashtree_add(the_hashtree, msgid);
    switch (artinfo->msgid->status)
    {
	case msgid_Virginal:
	    /* Only do history_check on virginal messages, because "check"
	     * method will tell us it's a dup if it gets checked more than once
	     * (took me ages to work that bug out!) */
	    if (history_check(msgid))
	    {
		deblog_debug(_("%s/%ld %s is history dup %ld\n"), group->name,
			artnum, msgid, ++hist_dups);
		artinfo->msgid->status = msgid_BeingFetched;
		artinfo->status = artnum_Duplicate;
		stats_inc_history_dups();
		return TRUE;
	    }
	    if (kill)
	    {
		stats_inc_killed();
		artinfo->status = artnum_Killed;
		hashtree_change_status(artinfo->msgid, msgid_Killed);
	    }
	    else
	    {
		stats_inc_total_to_fetch();
		groupcount_inc_articles(group->gc);
		artinfo->status = artnum_Available;
	    }
	    return FALSE;
	case msgid_Ready:
	    artinfo->status = kill ? artnum_Killed : artnum_Available;
	    if (kill)
	    {
		stats_dec_total_to_fetch();
		/* Killed article is quite likely to have been added from
		 * another group and killed due to crossposting, so probably
		 * better not to have such a function as
		 * groupcount_dec_articles */
		hashtree_change_status(artinfo->msgid, msgid_Killed);
	    }
	    return FALSE;
	case msgid_Killed:
	    artinfo->status = artnum_Killed;
	    return TRUE;
	default:
	    deblog_debug(_("%s/%d %s is a duplicate\n"), group->name, artnum,
		    msgid);
	    artinfo->status = artnum_Duplicate;
	    return TRUE;
    }
    return TRUE;
}

BOOL group_add_message_id(GroupData *group, long int artnum, const char *msgid)
{
    return group_add_or_kill_message_id(group, artnum, msgid, FALSE);
}

BOOL group_kill_message_id(GroupData *group, long int artnum, const char *msgid)
{
    return group_add_or_kill_message_id(group, artnum, msgid, TRUE);
}

long int group_get_next_article(GroupData *group, char const **msgid)
{
    long int art;

    deblog_debug(_("Rq for next article to download from %s\n"),
	    group->name);
    if (!msgid)
	deblog_debug(_("msgid is NULL, not committing download\n"));
    deblog_debug(_("Available: %ld-%ld, last requested = %ld\n"),
	    group->first_available, group->last_available,
	    group->last_requested);
    switch (group->status)
    {
	case group_Bogus:
	    deblog_info(_("Rq for article in bogus group %s\n"), group->name);
	    return 0;
	case group_AllDone:
	    deblog_debug(_("Group %s all done\n"), group->name);
	    return 0;
	case group_Downloading:
	    break;
	default:
	    deblog_info(_("Rq for article in group %s with no stats\n"),
		    group->name);
	    return -1;
    }
    for (art = group->last_requested + 1 - group->first_available; art <=
	    group->last_available - group->first_available; ++art)
    {
	deblog_debug(_("Looking at article %s/%ld (%ld)\n"),
		group->name, art, art + group->first_available);
	switch (group->artnums[art].status)
	{
	    case artnum_Available:
		if (msgid)
		    group->artnums[art].status = artnum_Touched;
		switch (group->artnums[art].msgid->status)
		{
		    case msgid_Virginal:
		    case msgid_Ready:
			deblog_debug(_("Requesting %s/%ld\n"), group->name,
				art + group->first_available);
			if (msgid)
			{
			    hashtree_change_status(group->artnums[art].msgid,
				    msgid_BeingFetched);
			    if ((group->last_requested = art +
					group->first_available) >=
				    group->last_available)
			    {
				deblog_debug(
					_("Last article in %s requested\n"),
					group->name);
				group->status = group_AllDone;
			    }
			    *msgid = group->artnums[art].msgid->msgid;
			}
			return group->last_requested;
		    case msgid_Fetched:
		    case msgid_BeingFetched:
			stats_inc_current_dups();
			/* Fall through */
		    case msgid_Killed:
			deblog_debug(_("%s/%ld %s already "
				    "killed/requested/downloaded\n"),
				group->name, art + group->first_available,
				group->artnums[art].msgid->msgid);
			/* stats_dec_total_to_fetch(); */
			/* This article must have been requested from a
			 * different group (x-posting) or different server:-
			 * need to make this group  behave as if it were
			 * requested from here; artnum_status can be left as
			 * Touched */
			group->artnums[art].status = artnum_Touched;
			group->last_requested = art + group->first_available;
			break;
		}
		break;
	    case artnum_Unknown:
		deblog_info(_("%s/%ld has unknown status at request\n"),
			group->name, art + group->first_available);
		/* This will probably be fetched later when info has overtaken
		 * requests again, so don't dec total */
		/* stats_dec_total_to_fetch(); */
		return -1;
	    case artnum_Unavailable:
		/* Unavailable articles have no known Message-ID */
		deblog_info(_("%s/%ld, is unavailable for request\n"),
			group->name, art + group->first_available);
		/*stats_dec_total_to_fetch();*/
		break;
	    case artnum_Duplicate:
		deblog_debug(_("%s/%ld %s is a duplicate\n"),
			group->name, art + group->first_available,
			group->artnums[art].msgid->msgid);
		break;
	    case artnum_Killed:
		deblog_debug(_("%s/%ld %s was killed\n"),
			group->name, art + group->first_available,
			group->artnums[art].msgid->msgid);
		break;
	    default:
		/* Only bugs in newsstar should be able to cause this */
		deblog_critical(_("Repeat request for %s/%ld %s\n"),
			group->name, art + group->first_available,
			group->artnums[art].msgid ?
			group->artnums[art].msgid->msgid : _("<unknown>"));
		/* stats_dec_total_to_fetch(); */
		break;
	}
    }
    if (art + group->first_available > group->last_available && msgid)
    {
	deblog_info(_("All articles in %s requested\n"), group->name);
	group->status = group_AllDone;
    }
    return 0;
}

void group_prune_unavailable(GroupData *group)
{
    long int art;

    for (art = 0; art <= group->last_available - group->first_available; ++art)
    {
	if (group->artnums[art].status == artnum_Unknown)
	{
	    deblog_debug(_("%s/%d is unavailable\n"), group->name, art +
		    group->first_available);
	    group->artnums[art].status = artnum_Unavailable;
	    group_add_backfill(group, art + group->first_available, TRUE);
	}
    }
}

void group_have_downloaded(GroupData *group, long int artnum)
{
    hashtree_change_status(group->artnums[artnum -
	    group->first_available].msgid, msgid_Fetched);
}

/* NB don't use deblog, can cause recursion trouble */
void group_write_newsrc(GroupData *group, FILE *fp)
{
    long int last = group->last_downloaded;

    /*
    if (group->status == group_Bogus)
    {
	if (main_normal_exit)
	{
	    fprintf(stderr,
		    _("Warning: Removing bogus group %s from newsrc.%s\n"),
		    group->name, group->serv->identifier);
	}
	return;
    }
    */
    /* Find last downloaded article */
    if (group->last_requested > 0)
    {
	BOOL found_last = FALSE;
	BOOL updated = FALSE;

	for (last = 0;
		last <= group->last_available - group->first_available
		&& !found_last;
		++last)
	{
	    switch (group->artnums[last].status)
	    {
		case artnum_Unknown:
		case artnum_Available:
		    found_last = TRUE;
		    break;
		case artnum_Touched:
		    updated = TRUE;
		    if (group->artnums[last].msgid->status != msgid_Fetched &&
			    group->artnums[last].msgid->status != msgid_Killed)
			found_last = TRUE;
		    break;
		default:
		/*
		case artnum_Unavailable:
		case artnum_Duplicate:
		case artnum_Killed:
		*/
		    updated = TRUE;
		    /* Skip to next article */
		    break;
	    }
	}
	if (updated)
	{
	    /* If we fell off end of loop last will be one more than it should
	     */
	    last += group->first_available;
	    if (last > group->last_available)
		last = group->last_available;
	}
	else
	{
	    last = group->last_downloaded;
	}

    }
    fprintf(fp, "%s %ld\n", group->name, last);
}

BackFillGroup *group_get_backfill(GroupData *group, BOOL create)
{
    if (!group->bfgroup)
    {
	if (group->serv->bfserv)
	{
	    deblog_debug(_("%s has backfills\n"), group->serv->identifier);
	    group->bfgroup = backfill_get_group(group->serv->bfserv,
		    group->name);
	    if (!group->bfgroup)
	    {
		if (create)
		{
		    deblog_debug(_("Creating backfills for %s\n"), group->name);
		    group->bfgroup = backfill_add_group(group->serv->bfserv,
			group->name);
		}
		else
		{
		    deblog_debug(_("No backfills for %s\n"), group->name);
		}
	    }
	    else
	    {
		deblog_debug(_("Adding backfills for %s\n"), group->name);
	    }
	}
    }
    return group->bfgroup;
}

void group_add_backfill(GroupData *group, long int artnum, BOOL no_xover)
{
    if (group->serv->bfserv)
    {
	backfill_add_unavailable(group_get_backfill(group, TRUE), artnum,
		no_xover);
    }
}

==> newsstar-1.5.3.tar.gz: newsstar-1.5.3/src/group.c <==
/*
    newsstar - advanced news fetcher
    Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "common.h"

#include "group.h"
#include "history.h"
#include "main.h"
#include "server.h"
#include "stats.h"

GroupData *group_new(const char *name, long int last_downloaded,
	ServerData *serv)
{
    GroupData *group = XMALLOC(sizeof(GroupData));

    deblog_info(_("Creating group %s on %s, last downloaded %ld\n"),
	    name, serv->identifier, last_downloaded);
    group->name = XSTRDUP(name);
    group->last_downloaded = last_downloaded;
    group->first_available = group->last_available = 0;
    group->last_requested = -1;
    group->status = group_Virginal;
    group->artnums = NULL;
    group->nthreads = 0;
    group->serv = serv;
    group->bfgroup = NULL;
    group->gc = groupcount_add_group(name);
    return group;
}

void group_delete(GroupData *group)
{
    deblog_debug(_("Deleting group %s\n"), group->name);
    groupcount_server_done(group->gc);
    free(group->artnums);
    free(group->name);
    free(group);
}

BOOL group_got_stats(GroupData *group, long int first, long int last,
	int thread_idx)
{
    int i;
    long int new_arts;

    deblog_debug(_("Got stats for %s on %s\n"),
	    group->name, group->serv->identifier);
    group->first_available = first;
    group->last_available = last;
    if (group->last_downloaded >= last)
    {
	group->status = group_AllDone;
	if (group->last_downloaded > last)
	{
	    deblog_warning(_("Last downloaded > last available in %s on %s\n"),
		    group->name, group->serv->identifier);
	    group->last_downloaded = last;
	}
	else
	{
	    deblog_message_thread(thread_idx, _("No new articles in %s\n"),
		    group->name);
	}
	return FALSE;
    }
    group->status = group_Downloading;
    if (group->last_downloaded < 0)
	group->last_downloaded = last + group->last_downloaded;
    if (group->last_downloaded < first)
	group->last_downloaded = first - 1;
    group->first_available = group->last_downloaded + 1;
    group->last_requested = group->first_available - 1;
    new_arts = last - group->first_available + 1;
    /* If this is a new group we may have slipped through above test for
     * 0 new articles so check again */
    if (!new_arts)
    {
	deblog_message_thread(thread_idx, _("No new articles in %s\n"),
		group->name);
	return FALSE;
    }
    deblog_message_thread(thread_idx, _("%ld new %s in %s\n"), new_arts,
	    new_arts == 1 ?  _("article") : _("articles"),
	    group->name);
    stats_inc_total_seen(new_arts);
    group->artnums = XMALLOC(new_arts * sizeof(group_artnum));
    for (i = 0; i < new_arts; ++i)
    {
	group->artnums[i].status = artnum_Unknown;
	group->artnums[i].msgid = NULL;
    }
    return TRUE;
}

static BOOL group_add_or_kill_message_id(GroupData *group, long int artnum,
	const char *msgid, BOOL kill)
{
    static long int hist_dups = 0;
    group_artnum *artinfo = &group->artnums[artnum - group->first_available];
    artinfo->msgid = hashtree_add(the_hashtree, msgid);
    switch (artinfo->msgid->status)
    {
	case msgid_Virginal:
	    /* Only do history_check on virginal messages, because "check"
	     * method will tell us it's a dup if it gets checked more than once
	     * (took me ages to work that bug out!) */
	    if (history_check(msgid))
	    {
		deblog_debug(_("%s/%ld %s is history dup %ld\n"), group->name,
			artnum, msgid, ++hist_dups);
		artinfo->msgid->status = msgid_BeingFetched;
		artinfo->status = artnum_Duplicate;
		stats_inc_history_dups();
		return TRUE;
	    }
	    if (kill)
	    {
		stats_inc_killed();
		artinfo->status = artnum_Killed;
		hashtree_change_status(artinfo->msgid, msgid_Killed);
	    }
	    else
	    {
		stats_inc_total_to_fetch();
		groupcount_inc_articles(group->gc);
		artinfo->status = artnum_Available;
	    }
	    return FALSE;
	case msgid_Ready:
	    artinfo->status = kill ? artnum_Killed : artnum_Available;
	    if (kill)
	    {
		stats_dec_total_to_fetch();
		/* Killed article is quite likely to have been added from
		 * another group and killed due to crossposting, so probably
		 * better not to have such a function as
		 * groupcount_dec_articles */
		hashtree_change_status(artinfo->msgid, msgid_Killed);
	    }
	    return FALSE;
	case msgid_Killed:
	    artinfo->status = artnum_Killed;
	    return TRUE;
	default:
	    deblog_debug(_("%s/%d %s is a duplicate\n"), group->name, artnum,
		    msgid);
	    artinfo->status = artnum_Duplicate;
	    return TRUE;
    }
    return TRUE;
}

BOOL group_add_message_id(GroupData *group, long int artnum, const char *msgid)
{
    return group_add_or_kill_message_id(group, artnum, msgid, FALSE);
}

BOOL group_kill_message_id(GroupData *group, long int artnum, const char *msgid)
{
    return group_add_or_kill_message_id(group, artnum, msgid, TRUE);
}

long int group_get_next_article(GroupData *group, char const **msgid)
{
    long int art;

    deblog_debug(_("Rq for next article to download from %s\n"),
	    group->name);
    if (!msgid)
	deblog_debug(_("msgid is NULL, not committing download\n"));
    deblog_debug(_("Available: %ld-%ld, last requested = %ld\n"),
	    group->first_available, group->last_available,
	    group->last_requested);
    switch (group->status)
    {
	case group_Bogus:
	    deblog_info(_("Rq for article in bogus group %s\n"), group->name);
	    return 0;
	case group_AllDone:
	    deblog_debug(_("Group %s all done\n"), group->name);
	    return 0;
	case group_Downloading:
	    break;
	default:
	    deblog_info(_("Rq for article in group %s with no stats\n"),
		    group->name);
	    return -1;
    }
    for (art = group->last_requested + 1 - group->first_available; art <=
	    group->last_available - group->first_available; ++art)
    {
	deblog_debug(_("Looking at article %s/%ld (%ld)\n"),
		group->name, art, art + group->first_available);
	switch (group->artnums[art].status)
	{
	    case artnum_Available:
		if (msgid)
		    group->artnums[art].status = artnum_Touched;
		switch (group->artnums[art].msgid->status)
		{
		    case msgid_Virginal:
		    case msgid_Ready:
			deblog_debug(_("Requesting %s/%ld\n"), group->name,
				art + group->first_available);
			if (msgid)
			{
			    hashtree_change_status(group->artnums[art].msgid,
				    msgid_BeingFetched);
			    if ((group->last_requested = art +
					group->first_available) >=
				    group->last_available)
			    {
				deblog_debug(
					_("Last article in %s requested\n"),
					group->name);
				group->status = group_AllDone;
			    }
			    *msgid = group->artnums[art].msgid->msgid;
			}
			return art + group->first_available;
		    case msgid_Fetched:
		    case msgid_BeingFetched:
			stats_inc_current_dups();
			/* Fall through */
		    case msgid_Killed:
			deblog_debug(_("%s/%ld %s already "
				    "killed/requested/downloaded\n"),
				group->name, art + group->first_available,
				group->artnums[art].msgid->msgid);
			/* stats_dec_total_to_fetch(); */
			/* This article must have been requested from a
			 * different group (x-posting) or different server:-
			 * need to make this group  behave as if it were
			 * requested from here; artnum_status can be left as
			 * Touched */
			group->artnums[art].status = artnum_Touched;
			group->last_requested = art + group->first_available;
			break;
		}
		break;
	    case artnum_Unknown:
		deblog_info(_("%s/%ld has unknown status at request\n"),
			group->name, art + group->first_available);
		/* This will probably be fetched later when info has overtaken
		 * requests again, so don't dec total */
		/* stats_dec_total_to_fetch(); */
		return -1;
	    case artnum_Unavailable:
		/* Unavailable articles have no known Message-ID */
		deblog_info(_("%s/%ld, is unavailable for request\n"),
			group->name, art + group->first_available);
		/*stats_dec_total_to_fetch();*/
		break;
	    case artnum_Duplicate:
		deblog_debug(_("%s/%ld %s is a duplicate\n"),
			group->name, art + group->first_available,
			group->artnums[art].msgid->msgid);
		break;
	    case artnum_Killed:
		deblog_debug(_("%s/%ld %s was killed\n"),
			group->name, art + group->first_available,
			group->artnums[art].msgid->msgid);
		break;
	    default:
		/* Only bugs in newsstar should be able to cause this */
		deblog_critical(_("Repeat request for %s/%ld %s\n"),
			group->name, art + group->first_available,
			group->artnums[art].msgid ?
			group->artnums[art].msgid->msgid : _("<unknown>"));
		/* stats_dec_total_to_fetch(); */
		break;
	}
    }
    if (art + group->first_available > group->last_available && msgid)
    {
	deblog_info(_("All articles in %s requested\n"), group->name);
	group->status = group_AllDone;
    }
    return 0;
}

void group_prune_unavailable(GroupData *group)
{
    long int art;

    for (art = 0; art <= group->last_available - group->first_available; ++art)
    {
	if (group->artnums[art].status == artnum_Unknown)
	{
	    deblog_debug(_("%s/%d is unavailable\n"), group->name, art +
		    group->first_available);
	    group->artnums[art].status = artnum_Unavailable;
	    group_add_backfill(group, art + group->first_available, TRUE);
	}
    }
}

void group_have_downloaded(GroupData *group, long int artnum)
{
    hashtree_change_status(group->artnums[artnum -
	    group->first_available].msgid, msgid_Fetched);
}

/* NB don't use deblog, can cause recursion trouble */
void group_write_newsrc(GroupData *group, FILE *fp)
{
    long int last = group->last_downloaded;

    /*
    if (group->status == group_Bogus)
    {
	if (main_normal_exit)
	{
	    fprintf(stderr,
		    _("Warning: Removing bogus group %s from newsrc.%s\n"),
		    group->name, group->serv->identifier);
	}
	return;
    }
    */
    /* Find last downloaded article */
    if (group->last_requested > 0)
    {
	BOOL found_last = FALSE;
	BOOL updated = FALSE;

	for (last = 0;
		last <= group->last_available - group->first_available
		&& !found_last;
		++last)
	{
	    switch (group->artnums[last].status)
	    {
		case artnum_Unknown:
		case artnum_Available:
		    found_last = TRUE;
		    break;
		case artnum_Touched:
		    updated = TRUE;
		    if (group->artnums[last].msgid->status != msgid_Fetched &&
			    group->artnums[last].msgid->status != msgid_Killed)
			found_last = TRUE;
		    break;
		default:
		/*
		case artnum_Unavailable:
		case artnum_Duplicate:
		case artnum_Killed:
		*/
		    updated = TRUE;
		    /* Skip to next article */
		    break;
	    }
	}
	if (updated)
	{
	    /* If we fell off end of loop last will be one more than it should
	     */
	    last += group->first_available;
	    if (last > group->last_available)
		last = group->last_available;
	}
	else
	{
	    last = group->last_downloaded;
	}

    }
    fprintf(fp, "%s %ld\n", group->name, last);
}

BackFillGroup *group_get_backfill(GroupData *group, BOOL create)
{
    if (!group->bfgroup)
    {
	if (group->serv->bfserv)
	{
	    deblog_debug(_("%s has backfills\n"), group->serv->identifier);
	    group->bfgroup = backfill_get_group(group->serv->bfserv,
		    group->name);
	    if (!group->bfgroup)
	    {
		if (create)
		{
		    deblog_debug(_("Creating backfills for %s\n"), group->name);
		    group->bfgroup = backfill_add_group(group->serv->bfserv,
			group->name);
		}
		else
		{
		    deblog_debug(_("No backfills for %s\n"), group->name);
		}
	    }
	    else
	    {
		deblog_debug(_("Adding backfills for %s\n"), group->name);
	    }
	}
    }
    return group->bfgroup;
}

void group_add_backfill(GroupData *group, long int artnum, BOOL no_xover)
{
    if (group->serv->bfserv)
    {
	backfill_add_unavailable(group_get_backfill(group, TRUE), artnum,
		no_xover);
    }
}
